//! The module for images that use [indexed
//! color](https://en.wikipedia.org/wiki/Indexed_color).
//!
//! Because the standard ops of Rust are insufficiently general, most of the
//! functionality actually comes from the use of the `ReadableIndexmap` and
//! `WritableIndexmap` traits. I know it's slightly inconvenient, so please
//! understand, I'm so sorry, and other apology memes.

use std::iter::*;
use std::marker::PhantomData;
use std::ops::*;

/// The generalized form of anything that you can read from as if it was part of
/// an `Indexmap`.
///
/// As with the `Iterator` trait, there's a few required methods you should
/// override if you want to implement this, and then a series of provided
/// methods that you should _really not_ override.
pub trait ReadableIndexmap: Index<(usize, usize), Output = u8> {
  /// The width of the image.
  fn width(&self) -> usize;

  /// The height of the image.
  fn height(&self) -> usize;

  /// The pitch between rows.
  ///
  /// This _might or might not be_ the same as the width, so you must use this
  /// value when jumping around with the pointer.
  fn pitch(&self) -> isize;

  /// The pointer to the start of the pixel data for this view.
  ///
  /// Obviously it's not actually unsafe to simply hand over a pixel pointer,
  /// but this is marked as unsafe because it's easy enough to forget that
  /// pointers don't track lifetimes, and so you better be careful.
  unsafe fn const_pixel_pointer(&self) -> *const u8;

  /// Grabs out a sub-slice of the data.
  ///
  /// If the requested area is out of range in any way, or if the requested area
  /// is less than 1 pixel for height or width, then you get `None` back. The
  /// range given is "sorted" so that you can request from lower left to upper
  /// right, or lower right to upper left, or however you like.
  ///
  /// ```rust
  /// use retro_pixel::indexmap::*;
  /// let im = Indexmap::new(10,15);
  ///
  /// let whole_slice = im.slice( (0,0) .. (im.width(),im.height()) ).unwrap();
  /// assert_eq!(whole_slice.width(), im.width());
  /// assert_eq!(whole_slice.height(), im.height());
  /// assert_eq!(whole_slice.pitch(), im.pitch());
  /// unsafe { assert_eq!(whole_slice.const_pixel_pointer(), im.const_pixel_pointer()) };
  ///
  /// let partial_slice = whole_slice.slice( (3,3) .. (7,8) ).unwrap();
  /// assert_eq!(partial_slice.width(), 4);
  /// assert_eq!(partial_slice.height(), 5);
  /// assert_eq!(partial_slice.pitch(), im.pitch());
  /// unsafe {
  ///   assert_eq!( partial_slice.const_pixel_pointer(),
  ///               im.const_pixel_pointer().offset((3 * im.pitch()) + 3) )
  /// };
  ///
  /// let no_slice = partial_slice.slice((2,2) .. (12,4));
  /// assert!(no_slice.is_none());
  /// ```
  fn slice<'a>(&'a self, r: Range<(usize, usize)>) -> Option<IndexmapRef<'a>> {
    let r = (r.start.0.min(r.end.0), r.start.1.min(r.end.1))..(r.start.0.max(r.end.0), r.start.1.max(r.end.1));
    if r.end.0 <= self.width() && r.end.1 <= self.height() && r.start.0 < r.end.0 && r.start.1 < r.end.1 {
      let width = r.end.0 - r.start.0;
      let height = r.end.1 - r.start.1;
      let ptr = unsafe {
        self
          .const_pixel_pointer()
          .offset((r.start.1 as isize * self.pitch()) + r.start.0 as isize)
      };
      let pitch = self.pitch();
      Some(IndexmapRef {
        width,
        height,
        ptr,
        pitch,
        _marker: PhantomData,
      })
    } else {
      None
    }
  }

  /// This is like `to_owned`, you get your own version of the data.
  ///
  /// The actual `ToOwned` trait has conflicts when you try to write a default
  /// impl for it based on `ReadableIndexmap`, so until there's a fix for that
  /// (specialization maybe?) we have this instead.
  ///
  /// ```rust
  /// use retro_pixel::indexmap::*;
  /// let mut source = Indexmap::new(5,5);
  /// let a: Indexmap = source.to_indexmap();
  /// let b: Indexmap = source.slice((0,0)..(5,5)).unwrap().to_indexmap();
  /// let c: Indexmap = source.slice_mut((0,0)..(5,5)).unwrap().to_indexmap();
  ///
  /// assert_eq!(a, b);
  /// assert_eq!(a, c);
  /// ```
  fn to_indexmap(&self) -> Indexmap
  where
    Self: Sized,
  {
    let mut output = Indexmap::new(self.width(), self.height());

    output.blit_from(self, (0, 0));

    output
  }

  /// Lets you iterate any readable form of Indexmap.
  fn iter(&self) -> IndexmapRefIter {
    IndexmapRefIter {
      ptr: unsafe { self.const_pixel_pointer() },
      next_x: 0,
      next_y: 0,
      width: self.width(),
      height: self.height(),
      pitch: self.pitch(),
      _marker: PhantomData,
    }
  }

  /// Scales up an indexmap by the given amount.
  ///
  /// ```rust
  /// use retro_pixel::indexmap::*;
  /// let mut source = Indexmap::new(2,2);
  /// source[(0,1)] = 5;
  ///
  /// let scaled = source.upscale(2);
  ///
  /// for (x,y,&p) in scaled.iter() {
  ///   match (x,y) {
  ///     (0,2) | (0,3) | (1,2) | (1,3) => assert_eq!(p, 5),
  ///     _ => assert_eq!(p, 0),
  ///   }
  /// }
  /// ```
  fn upscale(&self, scale: usize) -> Indexmap
  where
    Self: Index<(usize, usize), Output = u8>,
  {
    let mut output = Indexmap::new(self.width() * scale, self.height() * scale);

    for (x, y, &p) in self.iter() {
      for x_sub in 0..scale {
        for y_sub in 0..scale {
          output[(x * scale + x_sub, y * scale + y_sub)] = p;
        }
      }
    }

    output
  }
}

/// The generalized form of anything that you can write to as if it was part of
/// an `Indexmap`.
///
/// As with the `Iterator` trait, there's a few required methods you should
/// override if you want to implement this, and then a series of provided
/// methods that you should _really not_ override.
pub trait WritableIndexmap: ReadableIndexmap + IndexMut<(usize, usize), Output = u8> {
  /// The pointer to the start of the pixel data for this view.
  ///
  /// Obviously it's not actually unsafe to simply hand over a pixel pointer,
  /// but this is marked as unsafe because it's easy enough to forget that
  /// pointers don't track lifetimes, and so you better be careful.
  unsafe fn mut_pixel_pointer(&mut self) -> *mut u8;

  /// Grabs out a mutable sub-slice of the data.
  ///
  /// If the requested area is out of range in any way, or if the requested area
  /// is less than 1 pixel for height or width, then you get `None` back. The
  /// range given is "sorted" so that you can request from lower left to upper
  /// right, or lower right to upper left, or however you like.
  ///
  /// ```rust
  /// use retro_pixel::indexmap::*;
  /// let mut im = Indexmap::new(10,15);
  /// let im_mut_ptr = unsafe { im.mut_pixel_pointer() };
  ///
  /// let mut whole_slice = im.slice_mut( (0,0) .. (10,15) ).unwrap();
  /// assert_eq!(whole_slice.width(), 10);
  /// assert_eq!(whole_slice.height(), 15);
  /// assert_eq!(whole_slice.pitch(), 10);
  /// unsafe { assert_eq!(whole_slice.mut_pixel_pointer(), im_mut_ptr) };
  ///
  /// let mut partial_slice = whole_slice.slice_mut( (3,3) .. (7,8) ).unwrap();
  /// assert_eq!(partial_slice.width(), 4);
  /// assert_eq!(partial_slice.height(), 5);
  /// assert_eq!(partial_slice.pitch(), 10);
  /// unsafe {
  ///   assert_eq!( partial_slice.const_pixel_pointer(),
  ///               im_mut_ptr.offset((3 * 10) + 3) )
  /// };
  ///
  /// let no_slice = partial_slice.slice((2,2) .. (12,4));
  /// assert!(no_slice.is_none());
  /// ```
  fn slice_mut<'a>(&'a mut self, r: Range<(usize, usize)>) -> Option<IndexmapMut<'a>> {
    let r = (r.start.0.min(r.end.0), r.start.1.min(r.end.1))..(r.start.0.max(r.end.0), r.start.1.max(r.end.1));
    if r.end.0 <= self.width() && r.end.1 <= self.height() && r.start.0 < r.end.0 && r.start.1 < r.end.1 {
      let width = r.end.0 - r.start.0;
      let height = r.end.1 - r.start.1;
      let ptr = unsafe { self.mut_pixel_pointer().offset((r.start.1 as isize * self.pitch()) + r.start.0 as isize) };
      let pitch = self.pitch();
      Some(IndexmapMut {
        width,
        height,
        ptr,
        pitch,
        _marker: PhantomData,
      })
    } else {
      None
    }
  }

  /// Lets you mutably iterate over any writable form of indexmap.
  fn iter_mut(&mut self) -> IndexmapMutIter {
    IndexmapMutIter {
      ptr: unsafe { self.mut_pixel_pointer() },
      next_x: 0,
      next_y: 0,
      width: self.width(),
      height: self.height(),
      pitch: self.pitch(),
      _marker: PhantomData,
    }
  }

  /// Sets every position to be the given value.
  ///
  /// ```rust
  /// use retro_pixel::indexmap::*;
  /// let mut im = Indexmap::new(10,15);
  ///
  /// im.set_all(7);
  ///
  /// for y in 0 .. im.height() {
  ///   for x in 0 .. im.width() {
  ///     assert_eq!(im[(x,y)], 7);
  ///   }
  /// }
  /// ```
  fn set_all(&mut self, value: u8) {
    unsafe {
      let width = self.width();
      let height = self.height();
      let pitch = self.pitch();
      let mut ptr = self.mut_pixel_pointer();
      let mut y = 0;
      while y < height {
        ::std::ptr::write_bytes(ptr, value, width);
        ptr = ptr.offset(pitch);
        y += 1;
      }
    }
  }

  /// Copies data from the source indexmap into this indexmap.
  ///
  /// The source image is copied in at the offset given. This can even be a
  /// negative position if you want. The resulting region is automatically
  /// clipped to stay within bounds.
  ///
  /// Because of rust's lifetime rules this is able to use
  /// `copy_nonoverlapping`, so it's pretty zippy.
  ///
  /// ```rust
  /// use retro_pixel::indexmap::*;
  /// let mut dest = Indexmap::new(5,5);
  /// let mut src = Indexmap::new(2,3);
  /// src.set_all(3);
  /// src[(0,2)] = 5;
  /// src[(1,2)] = 5;
  ///
  /// dest.blit_from(&src, (1,-1));
  ///
  /// for y in 0 .. dest.height() {
  ///   for x in 0 .. dest.width() {
  ///     match (x,y) {
  ///       (1,0) | (2,0) => assert_eq!(dest[(x,y)], 3),
  ///       (1,1) | (2,1) => assert_eq!(dest[(x,y)], 5),
  ///       _ => assert_eq!(dest[(x,y)], 0),
  ///     }
  ///   }
  /// }
  /// ```
  fn blit_from<S>(&mut self, src: &S, offset: (isize, isize))
  where
    S: ReadableIndexmap,
  {
    let offset_x = offset.0;
    let offset_y = offset.1;
    let self_width = self.width() as isize;
    let self_height = self.height() as isize;
    let src_width = src.width() as isize;
    let src_height = src.height() as isize;
    // establish that we should be drawing something at all
    if offset_x < self_width && offset_y < self_height && -offset_x < src_width && -offset_y < src_height {
      // determine where we'll be copying
      let dest_start_x = (offset_x).max(0);
      let dest_start_y = (offset_y).max(0);
      let src_start_x = (-offset_x).max(0);
      let src_start_y = (-offset_y).max(0);
      let clip_width = (self_width - dest_start_x).min(src_width - src_start_x);
      let clip_height = (self_height - dest_start_y).min(src_height - src_start_y);
      assert!(clip_width > 0);
      assert!(clip_height > 0);
      unsafe {
        let dest_pitch = self.pitch();
        let mut dest_ptr = self.mut_pixel_pointer().offset(dest_start_y * dest_pitch + dest_start_x);
        let src_pitch = src.pitch();
        let mut src_ptr = src.const_pixel_pointer().offset(src_start_y * src_pitch + src_start_x);
        let clip_width = clip_width as usize;
        let clip_height = clip_height as usize;
        let mut y = 0;
        while y < clip_height {
          ::std::ptr::copy_nonoverlapping(src_ptr, dest_ptr, clip_width);
          src_ptr = src_ptr.offset(src_pitch);
          dest_ptr = dest_ptr.offset(dest_pitch);
          y += 1;
        }
      }
    }
  }

  /// As `blit_from`, but with any custom blit logic you want.
  ///
  /// The closure passed in will get the source value for the target location
  /// and the current value at the target location. It should return whatever
  /// value should end up in that location.
  ///
  /// ```rust
  /// use retro_pixel::indexmap::*;
  /// let mut dest = Indexmap::new(5,5);
  /// let mut src = Indexmap::new(2,3);
  /// dest.set_all(1);
  /// src.set_all(3);
  /// src[(0,2)] = 5;
  /// src[(1,2)] = 5;
  ///
  /// dest.blit_generic(&src, (1,-1), |src,dest| src+dest );
  ///
  /// for y in 0 .. dest.height() {
  ///   for x in 0 .. dest.width() {
  ///     match (x,y) {
  ///       (1,0) | (2,0) => assert_eq!(dest[(x,y)], 4),
  ///       (1,1) | (2,1) => assert_eq!(dest[(x,y)], 6),
  ///       _ => assert_eq!(dest[(x,y)], 1),
  ///     }
  ///   }
  /// }
  /// ```
  fn blit_generic<S, F>(&mut self, src: &S, offset: (isize, isize), op: F)
  where
    S: ReadableIndexmap,
    F: FnMut(u8, u8) -> u8,
  {
    let mut op = op;
    let offset_x = offset.0;
    let offset_y = offset.1;
    let self_width = self.width() as isize;
    let self_height = self.height() as isize;
    let src_width = src.width() as isize;
    let src_height = src.height() as isize;
    // establish that we should be drawing something at all
    if offset_x < self_width && offset_y < self_height && -offset_x < src_width && -offset_y < src_height {
      // determine where we'll be copying
      let dest_start_x = (offset_x).max(0);
      let dest_start_y = (offset_y).max(0);
      let src_start_x = (-offset_x).max(0);
      let src_start_y = (-offset_y).max(0);
      let clip_width = (self_width - dest_start_x).min(src_width - src_start_x);
      let clip_height = (self_height - dest_start_y).min(src_height - src_start_y);
      assert!(clip_width > 0);
      assert!(clip_height > 0);
      unsafe {
        let dest_pitch = self.pitch();
        let mut dest_ptr = self.mut_pixel_pointer().offset(dest_start_y * dest_pitch + dest_start_x);
        let src_pitch = src.pitch();
        let mut src_ptr = src.const_pixel_pointer().offset(src_start_y * src_pitch + src_start_x);
        let clip_width = clip_width as usize;
        let clip_height = clip_height as usize;
        let mut y = 0;
        while y < clip_height {
          let mut x = 0;
          let mut src_ptr_this_row = src_ptr;
          let mut dest_ptr_this_row = dest_ptr;
          while x < clip_width {
            *dest_ptr_this_row = op(*src_ptr_this_row, *dest_ptr_this_row);
            src_ptr_this_row = src_ptr_this_row.offset(1);
            dest_ptr_this_row = dest_ptr_this_row.offset(1);
            x += 1;
          }
          src_ptr = src_ptr.offset(src_pitch);
          dest_ptr = dest_ptr.offset(dest_pitch);
          y += 1;
        }
      }
    }
  }

  /// Flips all the rows of the `Indexmap`.
  ///
  /// ```rust
  /// use retro_pixel::indexmap::*;
  /// let mut im = Indexmap::new(2,2);
  /// im[(0,0)] = 5;
  /// im[(1,0)] = 5;
  ///
  /// im.flip_vertical();
  ///
  /// assert_eq!(im[(0,0)], 0);
  /// assert_eq!(im[(1,0)], 0);
  /// assert_eq!(im[(0,1)], 5);
  /// assert_eq!(im[(1,1)], 5);
  /// ```
  fn flip_vertical(&mut self) {
    unsafe {
      let swap_count = self.height() / 2;
      let pitch = self.pitch();
      let width = self.width();
      let mut low_ptr = self.mut_pixel_pointer();
      let mut high_ptr = self.mut_pixel_pointer().offset(pitch * (self.height() as isize - 1));
      let mut y = 0;
      let mut temp_buffer: Vec<u8> = vec![0; self.width()];
      while y < swap_count {
        let low_slice = ::std::slice::from_raw_parts_mut(low_ptr, width);
        let high_slice = ::std::slice::from_raw_parts_mut(high_ptr, width);
        temp_buffer.copy_from_slice(low_slice);
        low_slice.copy_from_slice(high_slice);
        high_slice.copy_from_slice(&temp_buffer);
        y += 1;
        low_ptr = low_ptr.offset(pitch);
        high_ptr = high_ptr.offset(-pitch);
      }
    }
  }

  /// Flips all the columns of the `Indexmap`.
  ///
  /// ```rust
  /// use retro_pixel::indexmap::*;
  /// let mut im = Indexmap::new(2,2);
  /// im[(0,0)] = 5;
  /// im[(0,1)] = 5;
  ///
  /// im.flip_horizontal();
  ///
  /// assert_eq!(im[(0,0)], 0);
  /// assert_eq!(im[(1,0)], 5);
  /// assert_eq!(im[(0,1)], 0);
  /// assert_eq!(im[(1,1)], 5);
  /// ```
  fn flip_horizontal(&mut self) {
    unsafe {
      let pitch = self.pitch();
      let width = self.width();
      let height = self.height();
      let mut ptr = self.mut_pixel_pointer();
      let mut y = 0;
      while y < height {
        ::std::slice::from_raw_parts_mut(ptr, width).reverse();
        y += 1;
        ptr = ptr.offset(pitch);
      }
    }
  }

  /// Rotates a square mutable slice 90 degrees counter-clockwise, in place.
  ///
  /// # Panics
  ///
  /// If the dimensions are not square this will panic.
  ///
  /// ```rust
  /// use retro_pixel::indexmap::*;
  /// let mut im = Indexmap::new(3,3);
  /// im[(0,0)] = 5;
  /// im[(2,0)] = 6;
  /// im[(2,2)] = 7;
  /// im[(0,2)] = 8;
  ///
  /// im.counterclockwise_rotate_square_inplace();
  ///
  /// assert_eq!(im[(0,0)], 8);
  /// assert_eq!(im[(2,0)], 5);
  /// assert_eq!(im[(2,2)], 6);
  /// assert_eq!(im[(0,2)], 7);
  /// ```
  fn counterclockwise_rotate_square_inplace(&mut self)
  where
    Self: Sized,
  {
    assert_eq!(self.width(), self.height());
    let n = self.width();
    let mut x = 0;
    while x < n / 2 {
      let mut y = x;
      while y < n - x - 1 {
        let temp = self[(x, y)];
        self[(x, y)] = self[(y, n - 1 - x)];
        self[(y, n - 1 - x)] = self[(n - 1 - x, n - 1 - y)];
        self[(n - 1 - x, n - 1 - y)] = self[(n - 1 - y, x)];
        self[(n - 1 - y, x)] = temp;
        //
        y += 1;
      }
      x += 1
    }
  }
}

/// This is the primary type, it actually owns the index data.
///
/// Almost no functionality exists on this type itself, instead you should use
/// it through the methods of the `ReadableIndexmap` and `WritableIndexmap`
/// traits.
#[derive(Debug, Clone, PartialEq, Eq)]
pub struct Indexmap {
  width: usize,
  height: usize,
  pixels: Vec<u8>,
}

impl Indexmap {
  /// Makes a new `Indexmap` with all locations as 0.
  ///
  /// As a sanity check, there's a `debug_assert!` that both dimensions given
  /// are less than the `isize` maximum.
  ///
  /// ```rust
  /// use retro_pixel::indexmap::*;
  /// let mut im = Indexmap::new(10,15);
  /// assert_eq!(im.width(), 10);
  /// assert_eq!(im.height(), 15);
  /// for y in 0 .. im.height() {
  ///   for x in 0 .. im.width() {
  ///     assert_eq!(im[(x,y)], 0);
  ///     im[(x,y)] = 7;
  ///     assert_eq!(im[(x,y)], 7);
  ///   }
  /// }
  /// ```
  pub fn new(width: usize, height: usize) -> Indexmap {
    debug_assert!(width < ::std::isize::MAX as usize);
    debug_assert!(height < ::std::isize::MAX as usize);
    Indexmap {
      width,
      height,
      pixels: vec![0; width * height],
    }
  }
}

impl IntoIterator for Indexmap {
  type Item = (usize, usize, u8);
  type IntoIter = IndexmapOwnedIter;

  /// ```rust
  /// use retro_pixel::indexmap::*;
  /// let image = Indexmap::new(10,10);
  /// let mut total = 0;
  /// for _pixel in image {
  ///   total += 1;
  /// }
  /// assert_eq!(total, 100);
  /// ```
  fn into_iter(self) -> IndexmapOwnedIter {
    let ptr = self.pixels.as_ptr();
    IndexmapOwnedIter {
      index: self,
      next_x: 0,
      next_y: 0,
      ptr,
    }
  }
}

impl Index<(usize, usize)> for Indexmap {
  type Output = u8;
  /// References the (x,y) location specified.
  ///
  /// # Panics
  ///
  /// * if `x` is greater than or equal to the width.
  /// * if `y` is greater than or equal to the height.
  fn index(&self, (x, y): (usize, usize)) -> &Self::Output {
    assert!(x < self.width);
    assert!(y < self.height);
    &self.pixels[y * self.width + x]
  }
}

#[test]
#[should_panic]
fn indexmap_index_impl_width_panic() {
  let im = Indexmap::new(10, 10);
  im[(10, 5)];
}

#[test]
#[should_panic]
fn indexmap_index_impl_height_panic() {
  let im = Indexmap::new(10, 10);
  im[(3, 10)];
}

impl ReadableIndexmap for Indexmap {
  fn width(&self) -> usize {
    self.width
  }

  fn height(&self) -> usize {
    self.height
  }

  fn pitch(&self) -> isize {
    self.width as isize
  }

  unsafe fn const_pixel_pointer(&self) -> *const u8 {
    self.pixels.as_ptr()
  }
}

impl IndexMut<(usize, usize)> for Indexmap {
  /// Mut references the (x,y) location specified.
  ///
  /// # Panics
  ///
  /// * if `x` is greater than or equal to the width.
  /// * if `y` is greater than or equal to the height.
  fn index_mut(&mut self, (x, y): (usize, usize)) -> &mut Self::Output {
    assert!(x < self.width);
    assert!(y < self.height);
    &mut self.pixels[y * self.width + x]
  }
}

#[test]
#[should_panic]
fn indexmap_index_mut_impl_width_panic() {
  let mut im = Indexmap::new(10, 10);
  im[(10, 5)] = 7;
}

#[test]
#[should_panic]
fn indexmap_index_mut_impl_height_panic() {
  let mut im = Indexmap::new(10, 10);
  im[(3, 10)] = 7;
}

impl WritableIndexmap for Indexmap {
  unsafe fn mut_pixel_pointer(&mut self) -> *mut u8 {
    self.pixels.as_mut_ptr()
  }
}

/// This is a read-only view of some Indexmap data.
///
/// Almost no functionality exists on this type itself, instead you should use
/// it through the methods of the `ReadableIndexmap` trait.
#[derive(Debug, Clone, PartialEq, Eq)]
pub struct IndexmapRef<'a> {
  width: usize,
  height: usize,
  pitch: isize,
  ptr: *const u8,
  _marker: PhantomData<&'a u8>,
}

impl<'a> Index<(usize, usize)> for IndexmapRef<'a> {
  type Output = u8;
  /// References the (x,y) location specified.
  ///
  /// # Panics
  ///
  /// * if `x` is greater than or equal to the width.
  /// * if `y` is greater than or equal to the height.
  ///
  /// ```rust
  /// use retro_pixel::indexmap::*;
  /// let im = Indexmap::new(10,15);
  /// let slice: IndexmapRef = im.slice( (0,0) .. (10,15) ).unwrap();
  /// assert_eq!(slice.width(), 10);
  /// assert_eq!(slice.height(), 15);
  /// for y in 0 .. slice.height() {
  ///   for x in 0 .. slice.width() {
  ///     assert_eq!(slice[(x,y)], 0);
  ///   }
  /// }
  /// ```
  fn index(&self, (x, y): (usize, usize)) -> &Self::Output {
    assert!(x < self.width);
    assert!(y < self.height);
    unsafe { self.ptr.offset((y as isize * self.pitch) + x as isize).as_ref().unwrap() }
  }
}

impl<'a> IntoIterator for IndexmapRef<'a> {
  type Item = (usize, usize, &'a u8);
  type IntoIter = IndexmapRefIter<'a>;

  fn into_iter(self) -> IndexmapRefIter<'a> {
    IndexmapRefIter {
      ptr: self.ptr,
      next_x: 0,
      next_y: 0,
      width: self.width,
      height: self.height,
      pitch: self.pitch,
      _marker: PhantomData,
    }
  }
}

#[test]
#[should_panic]
fn indexmap_ref_index_impl_width_panic() {
  let im = Indexmap::new(10, 10);
  let slice = im.slice((0, 0)..(10, 10)).unwrap();
  slice[(10, 5)];
}

#[test]
#[should_panic]
fn indexmap_ref_index_impl_height_panic() {
  let im = Indexmap::new(10, 10);
  let slice = im.slice((0, 0)..(10, 10)).unwrap();
  slice[(3, 10)];
}

impl<'a> ReadableIndexmap for IndexmapRef<'a> {
  fn width(&self) -> usize {
    self.width
  }

  fn height(&self) -> usize {
    self.height
  }

  fn pitch(&self) -> isize {
    self.pitch
  }

  unsafe fn const_pixel_pointer(&self) -> *const u8 {
    self.ptr
  }
}

/// This is a read/write view of some Indexmap data.
///
/// Almost no functionality exists on this type itself, instead you should use
/// it through the methods of the `ReadableIndexmap` and `WritableIndexmap`
/// traits.
#[derive(Debug, Clone, PartialEq, Eq)]
pub struct IndexmapMut<'a> {
  width: usize,
  height: usize,
  pitch: isize,
  ptr: *mut u8,
  _marker: PhantomData<&'a mut u8>,
}

impl<'a> Index<(usize, usize)> for IndexmapMut<'a> {
  type Output = u8;
  /// References the (x,y) location specified.
  ///
  /// # Panics
  ///
  /// * if `x` is greater than or equal to the width.
  /// * if `y` is greater than or equal to the height.
  ///
  /// ```rust
  /// use retro_pixel::indexmap::*;
  /// let mut im = Indexmap::new(10,15);
  /// let slice_mut: IndexmapMut = im.slice_mut( (0,0) .. (10,15) ).unwrap();
  /// assert_eq!(slice_mut.width(), 10);
  /// assert_eq!(slice_mut.height(), 15);
  /// for y in 0 .. slice_mut.height() {
  ///   for x in 0 .. slice_mut.width() {
  ///     assert_eq!(slice_mut[(x,y)], 0);
  ///   }
  /// }
  /// ```
  fn index(&self, (x, y): (usize, usize)) -> &Self::Output {
    assert!(x < self.width);
    assert!(y < self.height);
    unsafe { self.ptr.offset((y as isize * self.pitch) + x as isize).as_ref().unwrap() }
  }
}

impl<'a> IntoIterator for IndexmapMut<'a> {
  type Item = (usize, usize, &'a mut u8);
  type IntoIter = IndexmapMutIter<'a>;

  fn into_iter(self) -> IndexmapMutIter<'a> {
    IndexmapMutIter {
      ptr: self.ptr,
      next_x: 0,
      next_y: 0,
      width: self.width,
      height: self.height,
      pitch: self.pitch,
      _marker: PhantomData,
    }
  }
}

#[test]
#[should_panic]
fn indexmap_mut_index_impl_width_panic() {
  let mut im = Indexmap::new(10, 10);
  let slice_mut = im.slice_mut((0, 0)..(10, 10)).unwrap();
  slice_mut[(10, 5)];
}

#[test]
#[should_panic]
fn indexmap_mut_index_impl_height_panic() {
  let mut im = Indexmap::new(10, 10);
  let slice_mut = im.slice_mut((0, 0)..(10, 10)).unwrap();
  slice_mut[(3, 10)];
}

impl<'a> ReadableIndexmap for IndexmapMut<'a> {
  fn width(&self) -> usize {
    self.width
  }

  fn height(&self) -> usize {
    self.height
  }

  fn pitch(&self) -> isize {
    self.pitch
  }

  unsafe fn const_pixel_pointer(&self) -> *const u8 {
    self.ptr
  }
}

impl<'a> IndexMut<(usize, usize)> for IndexmapMut<'a> {
  /// Mut references the (x,y) location specified.
  ///
  /// # Panics
  ///
  /// * if `x` is greater than or equal to the width.
  /// * if `y` is greater than or equal to the height.
  ///
  /// ```rust
  /// use retro_pixel::indexmap::*;
  /// let mut im = Indexmap::new(10,15);
  /// let mut slice_mut: IndexmapMut = im.slice_mut( (0,0) .. (10,15) ).unwrap();
  /// assert_eq!(slice_mut.width(), 10);
  /// assert_eq!(slice_mut.height(), 15);
  /// for y in 0 .. slice_mut.height() {
  ///   for x in 0 .. slice_mut.width() {
  ///     assert_eq!(slice_mut[(x,y)], 0);
  ///     slice_mut[(x,y)] = 7;
  ///     assert_eq!(slice_mut[(x,y)], 7);
  ///   }
  /// }
  /// ```
  fn index_mut(&mut self, (x, y): (usize, usize)) -> &mut Self::Output {
    assert!(x < self.width);
    assert!(y < self.height);
    unsafe { self.ptr.offset((y as isize * self.pitch) + x as isize).as_mut().unwrap() }
  }
}

#[test]
#[should_panic]
fn indexmap_mut_index_mut_impl_width_panic() {
  let mut im = Indexmap::new(10, 10);
  let mut slice_mut = im.slice_mut((0, 0)..(10, 10)).unwrap();
  slice_mut[(10, 5)] = 7;
}

#[test]
#[should_panic]
fn indexmap_mut_index_mut_impl_height_panic() {
  let mut im = Indexmap::new(10, 10);
  let mut slice_mut = im.slice_mut((0, 0)..(10, 10)).unwrap();
  slice_mut[(3, 10)] = 7;
}

impl<'a> WritableIndexmap for IndexmapMut<'a> {
  unsafe fn mut_pixel_pointer(&mut self) -> *mut u8 {
    self.ptr
  }
}

/// A struct that iterates over the "owned" value of an `Indexmap`.
///
/// But they're `u8`s, so they're all copy, so yeah.
#[derive(Debug, PartialEq, Eq)]
pub struct IndexmapOwnedIter {
  index: Indexmap,
  next_x: usize,
  next_y: usize,
  ptr: *const u8,
}

impl Iterator for IndexmapOwnedIter {
  /// (x,y,index_value)
  type Item = (usize, usize, u8);

  /// ```rust
  /// use retro_pixel::indexmap::*;
  /// let mut im = Indexmap::new(2,2);
  /// im[(0,0)] = 1;
  /// im[(1,0)] = 2;
  /// im[(0,1)] = 3;
  /// im[(1,1)] = 4;
  /// let mut the_iter = im.into_iter();
  /// assert_eq!(the_iter.size_hint(), (4, Some(4)));
  /// assert_eq!(the_iter.next(), Some((0,0,1)));
  /// assert_eq!(the_iter.next(), Some((1,0,2)));
  /// assert_eq!(the_iter.next(), Some((0,1,3)));
  /// assert_eq!(the_iter.next(), Some((1,1,4)));
  /// assert_eq!(the_iter.next(), None);
  /// assert_eq!(the_iter.size_hint(), (0, Some(0)));
  /// ```
  fn next(&mut self) -> Option<Self::Item> {
    if self.next_y < self.index.height() {
      let out = Some((self.next_x, self.next_y, unsafe { *self.ptr }));
      self.ptr = unsafe { self.ptr.offset(1) };
      self.next_x += 1;
      if self.next_x == self.index.width() {
        self.next_x = 0;
        self.next_y += 1;
      }
      out
    } else {
      None
    }
  }

  fn size_hint(&self) -> (usize, Option<usize>) {
    let total = self.index.width() * self.index.height();
    let used = self.next_y * self.index.width() + self.next_x;
    let remaining = total.saturating_sub(used);
    (remaining, Some(remaining))
  }
}

/// A struct that iterates over referenced values of an `Indexmap`.
///
/// But they're `u8`s, so they're all copy, so yeah.
#[derive(Debug, PartialEq, Eq)]
pub struct IndexmapRefIter<'a> {
  ptr: *const u8,
  next_x: usize,
  next_y: usize,
  width: usize,
  height: usize,
  pitch: isize,
  _marker: PhantomData<&'a u8>,
}

impl<'a> Iterator for IndexmapRefIter<'a> {
  /// (x,y,index_value)
  type Item = (usize, usize, &'a u8);

  /// ```rust
  /// use retro_pixel::indexmap::*;
  /// let mut im = Indexmap::new(2,2);
  /// im[(0,0)] = 1;
  /// im[(1,0)] = 2;
  /// im[(0,1)] = 3;
  /// im[(1,1)] = 4;
  /// let mut the_iter = im.iter();
  /// assert_eq!(the_iter.size_hint(), (4, Some(4)));
  /// assert_eq!(the_iter.next(), Some((0,0,&1)));
  /// assert_eq!(the_iter.next(), Some((1,0,&2)));
  /// assert_eq!(the_iter.next(), Some((0,1,&3)));
  /// assert_eq!(the_iter.next(), Some((1,1,&4)));
  /// assert_eq!(the_iter.next(), None);
  /// assert_eq!(the_iter.size_hint(), (0, Some(0)));
  /// ```
  fn next(&mut self) -> Option<Self::Item> {
    if self.next_y < self.height {
      let out = Some((self.next_x, self.next_y, unsafe { self.ptr.as_ref().unwrap() }));
      self.ptr = unsafe { self.ptr.offset(1) };
      self.next_x += 1;
      if self.next_x == self.width {
        self.ptr = unsafe { self.ptr.offset(self.pitch - (self.width as isize)) };
        self.next_x = 0;
        self.next_y += 1;
      }
      out
    } else {
      None
    }
  }

  fn size_hint(&self) -> (usize, Option<usize>) {
    let total = self.width * self.height;
    let used = self.next_y * self.width + self.next_x;
    let remaining = total.saturating_sub(used);
    (remaining, Some(remaining))
  }
}

/// A struct that iterates over mut referenced values of an `Indexmap`.
#[derive(Debug, PartialEq, Eq)]
pub struct IndexmapMutIter<'a> {
  ptr: *mut u8,
  next_x: usize,
  next_y: usize,
  width: usize,
  height: usize,
  pitch: isize,
  _marker: PhantomData<&'a mut u8>,
}

impl<'a> Iterator for IndexmapMutIter<'a> {
  /// (x,y,index_value)
  type Item = (usize, usize, &'a mut u8);

  /// ```rust
  /// use retro_pixel::indexmap::*;
  /// let mut im = Indexmap::new(2,2);
  /// im[(0,0)] = 1;
  /// im[(1,0)] = 2;
  /// im[(0,1)] = 3;
  /// im[(1,1)] = 4;
  /// let mut the_iter = im.iter_mut();
  /// assert_eq!(the_iter.size_hint(), (4, Some(4)));
  /// assert_eq!(the_iter.next(), Some((0,0,&mut 1)));
  /// assert_eq!(the_iter.next(), Some((1,0,&mut 2)));
  /// assert_eq!(the_iter.next(), Some((0,1,&mut 3)));
  /// assert_eq!(the_iter.next(), Some((1,1,&mut 4)));
  /// assert_eq!(the_iter.next(), None);
  /// assert_eq!(the_iter.size_hint(), (0, Some(0)));
  /// ```
  fn next(&mut self) -> Option<Self::Item> {
    if self.next_y < self.height {
      let out = Some((self.next_x, self.next_y, unsafe { self.ptr.as_mut().unwrap() }));
      self.ptr = unsafe { self.ptr.offset(1) };
      self.next_x += 1;
      if self.next_x == self.width {
        self.ptr = unsafe { self.ptr.offset(self.pitch - (self.width as isize)) };
        self.next_x = 0;
        self.next_y += 1;
      }
      out
    } else {
      None
    }
  }

  fn size_hint(&self) -> (usize, Option<usize>) {
    let total = self.width * self.height;
    let used = self.next_y * self.width + self.next_x;
    let remaining = total.saturating_sub(used);
    (remaining, Some(remaining))
  }
}
